
//***************************************************
// Generalised Bitmap Module C++ adapter for GBM plugins
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
// Copyright (C) 2006-2007 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
//***************************************************

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include "os2defs.h"

#include <string.h>
#include <stdlib.h>

#include "GbmRenderer.hpp"

#include "gbm.h"
#include "gbmscale.h"
#include "gbmmir.h"
#include "gbmrect.h"

// ---------------------------------------------------------

GbmRenderer::GbmRenderer()
    : renderPage(NULL),
      unscaledWidth(0),
      unscaledHeight(0),
      mScaleFilter(GBM_SCALE_FILTER_MITCHELL),
      mHasScaleFilter(TRUE)
{
}

// ---------------------------------------------------------

GbmRenderer::~GbmRenderer()
{
    renderPage = NULL;
}

// ---------------------------------------------------------

void GbmRenderer::setRenderPage(GbmDocumentPage * page)
{
    if (page != renderPage)
    {
        // clear old render cache
        cacheDescription.clear();
        renderPage = page;
    }
}

GbmDocumentPage * GbmRenderer::getRenderPage() const
{
    return renderPage;
}

// ---------------------------------------------------------

void GbmRenderer::setScaleFilter(const GBM_SCALE_FILTER filter)
{
    if ((! mHasScaleFilter) ||
        (mHasScaleFilter && (mScaleFilter != filter)))
    {
       // clear old render cache to force recalculation
       cacheDescription.clear();
    }
    mScaleFilter    = filter;
    mHasScaleFilter = TRUE;
}

void GbmRenderer::removeScaleFilter()
{
    if (mHasScaleFilter)
    {
       // clear old render cache to force recalculation
       cacheDescription.clear();
    }
    mScaleFilter    = GBM_SCALE_FILTER_MITCHELL;
    mHasScaleFilter = FALSE;
}

// ---------------------------------------------------------

/**
 * @param src_x         x coordinate of upper left corner
 * @param src_y         y coordinate of upper left corner
 * @param src_width     width of rectangle to render
 * @param src_height    height of rectangle to render
 * @param scale_x       scale specified as pixels per point (width)
 * @param scale_y       scale specified as pixels per point (height)
 * @param renderBuffer  render buffer (4 byte aligned row size according
 *                      to src_width and page colour depth of page set)
 * @param bmpHeader     the bitmapinfo for direct use by GPI
 *
 * 1. Scale the document to match the specified pixels per point
 * 2. Render the rectangle given by the upper left corner at
 *    (src_x, src_y) and src_width and src_height.
 **/

void GbmRenderer::renderToBuffer(long        src_x    , long   src_y,
                                 long        src_width, long   src_height,
                                 double      scale_x  , double scale_y,
                                 byte      * renderBuffer,
                                 BMPHEADER * bmpHeader) throw ( GbmException )
{
    if (renderPage == NULL)
    {
        throw GbmException(GBM_ERR_BAD_ARG, "No render page set");
    }

    if ((scale_x < 0.0) || (scale_y < 0.0) ||
        (src_y < 0)     || (src_x < 0)     || (src_width <= 0) || (src_height <= 0))
    {
        throw GbmException(GBM_ERR_BAD_ARG, "Bad render options");
    }
    if (bmpHeader == NULL)
    {
        throw GbmException(GBM_ERR_BAD_ARG, "Missing bitmap header");
    }

    GBMRGB gbmrgb[0x100] = { 0 };
    BOOLEAN isGray8bpp   = FALSE;
    copyBmpHeaderAndGetPaletteAsGBMRGB(bmpHeader, gbmrgb, &isGray8bpp);

    // Check if any important render parameters (scaling, page size) have changed.
    if ((scale_x                 != cacheDescription.scale_x) ||
        (scale_y                 != cacheDescription.scale_y) ||
        (renderPage->getWidth()  != unscaledWidth)            ||
        (renderPage->getHeight() != unscaledHeight))
    {
        // force full recalculation of target bitmap
        cacheDescription.clear();
        unscaledWidth  = renderPage->getWidth();
        unscaledHeight = renderPage->getHeight();
    }

          int    copySrcWidth  = cacheDescription.width;
          int    copySrcHeight = cacheDescription.height;
          int    copySrcStride = cacheDescription.rowStride;
    const byte * copySrcData   = cacheDescription.getDataPtr();
          byte * scaledData    = NULL;

    if (copySrcData == NULL)
    {
        // the page is already rotate if this was necessary
        copySrcWidth  = renderPage->getWidth();
        copySrcHeight = renderPage->getHeight();
        copySrcStride = renderPage->getBitmapDataRowSize();
        copySrcData   = renderPage->getBitmapDataPtr();

        // Needs additional scaling ?

        // check what is more memory conservative (scale first or rotate first)
        if ((scale_x != 1.0) || (scale_y != 1.0))
        {
            const int scaledWidth  = (int) ((scale_x * ((double) (copySrcWidth ))) + 0.5);
            const int scaledHeight = (int) ((scale_y * ((double) (copySrcHeight))) + 0.5);
            const int scaledStride = ((scaledWidth * renderPage->getColorDepth() + 31)/32) * 4;

            scaledData = new byte[scaledStride * scaledHeight];
            if (scaledData != NULL)
            {
                GBM_ERR gbmrc = GBM_ERR_OK;

                // scale to requested size
                if (mHasScaleFilter && (isGray8bpp || (renderPage->getColorDepth() > 8)))
                {
                    if (isGray8bpp)
                    {
                        gbmrc = gbm_quality_scale_gray(copySrcData, copySrcWidth  , copySrcHeight,
                                                       renderPage->getColorDepth(), gbmrgb,
                                                       scaledData , scaledWidth   , scaledHeight,
                                                       gbmrgb     , mScaleFilter);
                        // update the color palette to be returned
                        copyGBMRGB2BmpHeader(gbmrgb, bmpHeader);
                    }
                    else
                    {
                        gbmrc = gbm_quality_scale_bgra(copySrcData, copySrcWidth  , copySrcHeight,
                                                       scaledData , scaledWidth   , scaledHeight,
                                                       renderPage->getColorDepth(), mScaleFilter);
                    }
                }
                else
                {
                    gbmrc = gbm_simple_scale(copySrcData, copySrcWidth , copySrcHeight,
                                             scaledData , scaledWidth  , scaledHeight,
                                             renderPage->getColorDepth());
                }
                if (gbmrc == GBM_ERR_OK)
                {
                    // set the copy source to the scaled data
                    copySrcWidth  = scaledWidth;
                    copySrcHeight = scaledHeight;
                    copySrcStride = scaledStride;
                    copySrcData   = scaledData;
                }
                else
                {
                    delete [] scaledData;
                    scaledData = NULL;
                    throw GbmException(gbmrc);
                }
            }
            else
            {
                /* In case we had a problem with scaling (e.g. out of memory), */
                /* render nothing (or later show an error icon?).              */
                throw GbmException(GBM_ERR_MEM);
            }
        }

        cacheDescription.width     = copySrcWidth;
        cacheDescription.height    = copySrcHeight;
        cacheDescription.rowStride = copySrcStride;
        cacheDescription.scale_x   = scale_x;
        cacheDescription.scale_y   = scale_y;

        if (copySrcData == scaledData)
        {
            // this data can be freed as we own it
            cacheDescription.setData((byte *) scaledData);
        }
        else
        {
            // just store the reference which will not be freed
            cacheDescription.setConstData(copySrcData);

            // Don't free the scaledData buffer if it points
            // to the same address as the cache.
            delete [] scaledData;
            scaledData = NULL;
        }
    }

    // copy visible rectangle to the drawing buffer
    GBM renderGbm;
    renderGbm.w   = cacheDescription.width;
    renderGbm.h   = cacheDescription.height;
    renderGbm.bpp = renderPage->getColorDepth();

    // protect against buffer overrun due to rounding errors
    if (src_x + src_width > renderGbm.w)
    {
        src_width = renderGbm.w - src_x;
    }
    if (src_y + src_height > renderGbm.h)
    {
        src_height = renderGbm.h - src_y;
    }

    gbm_subrectangle(&renderGbm,
                      src_x, renderGbm.h-(src_y+src_height), src_width, src_height,
                      cacheDescription.getDataPtr(), renderBuffer);
}

// ---------------------------------------------------------

/**
 * @param src_x           x coordinate of upper left corner
 * @param src_y           y coordinate of upper left corner
 * @param src_width       width of rectangle to render
 * @param src_height      height of rectangle to render
 * @param scale_x         scale specified as pixels per point (width)
 * @param scale_y         scale specified as pixels per point (height)
 * @param renderBuffer24  render buffer (4 byte aligned row size according
 *                        to src_width and 24bpp colour depth)
 * @param bmpHeader       the bitmapinfo for direct use by GPI
 *
 * 1. Scale the document to match the specified pixels per point
 * 2. Render the rectangle given by the upper left corner at
 *    (src_x, src_y) and src_width and src_height.
 **/

void GbmRenderer::renderToBuffer24bpp(long        src_x    , long   src_y,
                                      long        src_width, long   src_height,
                                      double      scale_x  , double scale_y,
                                      byte      * renderBuffer24,
                                      BMPHEADER * bmpHeader) throw ( GbmException )
{
    if (renderPage == NULL)
    {
        throw GbmException(GBM_ERR_BAD_ARG, "No render page set");
    }

    // if the source is already 24bpp, just forward to direct render function
    if (renderPage->getColorDepth() == 24)
    {
        renderToBuffer(src_x         , src_y,
                       src_width     , src_height,
                       scale_x       , scale_y,
                       renderBuffer24, bmpHeader);
        return;
    }

    if ((scale_x < 0.0) || (scale_y < 0.0) ||
        (src_y < 0)     || (src_x < 0)     || (src_width <= 0) || (src_height <= 0))
    {
        throw GbmException(GBM_ERR_BAD_ARG, "Bad render options");
    }

    const int scaledWidth  = (int) ((scale_x * ((double) (renderPage->getWidth()))) + 0.5);
    const int scaledHeight = (int) ((scale_y * ((double) (renderPage->getHeight()))) + 0.5);
    const int scaledStride = ((scaledWidth * renderPage->getColorDepth() + 31)/32) * 4;

    byte * scaledData = scaledData = new byte[scaledStride * scaledHeight];
    if (scaledData != NULL)
    {
        // first prepare the bitmap as usual but then expand it to 24bpp
        try
        {
            renderToBuffer(src_x     , src_y,
                           src_width , src_height,
                           scale_x   , scale_y,
                           scaledData, bmpHeader);

            expandTo24bpp(src_width, src_height,
                          renderPage->getColorDepth(),
                          scaledData, renderBuffer24);

            /* update relevant bitmap header data */
            const int bmpDataStride = ((src_width * 24 + 31)/32) * 4;
            const int bmpDataLength = bmpDataStride * src_height;

            bmpHeader->bmp2.cBitCount = 24;            /*  Number of bits per pel within a plane. */
            bmpHeader->bmp2.cbImage   = bmpDataLength; /*  Length of bit-map storage data, in bytes. */

        }
        catch(GbmException & ex)
        {
            delete [] scaledData;
            scaledData = NULL;
            throw ex;
        }
        delete [] scaledData;
        scaledData = NULL;
        // DONE
    }
    else
    {
        /* In case we had a problem with scaling (e.g. out of memory), */
        /* render nothing (or later show an error icon?).              */
        throw GbmException(GBM_ERR_MEM);
    }
}

// ---------------------------------------------------------

void GbmRenderer::expandTo24bpp(int src_w, int src_h, int src_bpp,
                                const byte *data, byte * data24) const throw ( GbmException )
{
    if ((src_bpp >= 24) || (data == NULL) || (data24 == NULL) || (renderPage == NULL) ||
        (src_w < 1) || (src_h < 1))
    {
        throw GbmException(GBM_ERR_BAD_ARG);
    }

    const BITMAPINFO2 * bmpInfo = renderPage->getBitmapInfoPtr();
    if (bmpInfo == NULL)
    {
        throw GbmException(GBM_ERR_NOT_SUPP);
    }

    const RGB2 * pRGB = (const RGB2 *)((const byte *)bmpInfo + bmpInfo->cbFix);
    if (pRGB == NULL)
    {
        throw GbmException(GBM_ERR_NOT_SUPP);
    }

    const int stride     = ((src_w * src_bpp + 31)/32) * 4;
    const int new_stride = ((src_w * 3 + 3) & ~3);
          int x, y;

    for ( y = 0; y < src_h; y++ )
    {
        const byte *src  = data + y * stride;
              byte *dest = data24 + y * new_stride;

        switch ( src_bpp )
        {
            case 1:
            {
                byte c;

                for ( x = 0; x < src_w; x++ )
                {
                    if ( (x & 7) == 0 )
                    {
                        c = *src++;
                    }
                    else
                    {
                        c <<= 1;
                    }
                    *dest++ = pRGB[c >> 7].bBlue;
                    *dest++ = pRGB[c >> 7].bGreen;
                    *dest++ = pRGB[c >> 7].bRed;
                }
            }
            break;

            case 4:
            {
                for ( x = 0; x + 1 < src_w; x += 2 )
                {
                    const byte c = *src++;

                    *dest++ = pRGB[c >> 4].bBlue;
                    *dest++ = pRGB[c >> 4].bGreen;
                    *dest++ = pRGB[c >> 4].bRed;
                    *dest++ = pRGB[c & 15].bBlue;
                    *dest++ = pRGB[c & 15].bGreen;
                    *dest++ = pRGB[c & 15].bRed;
                }

                if ( x < src_w )
                {
                    const byte c = *src;

                    *dest++ = pRGB[c >> 4].bBlue;
                    *dest++ = pRGB[c >> 4].bGreen;
                    *dest++ = pRGB[c >> 4].bRed;
                }
            }
            break;

            case 8:
            {
                for ( x = 0; x < src_w; x++ )
                {
                    const byte c = *src++;

                    *dest++ = pRGB[c].bBlue;
                    *dest++ = pRGB[c].bGreen;
                    *dest++ = pRGB[c].bRed;
                }
            }
            break;

            default:
               throw GbmException(GBM_ERR_NOT_SUPP);
        }
    }
}


/* ---------------------------- */

void GbmRenderer::copyBmpHeaderAndGetPaletteAsGBMRGB(BMPHEADER *bmpHeader,
                                                     GBMRGB    *gbmrgb,
                                                     BOOLEAN   *isGrayscale8bpp)
                                                         const throw ( GbmException )
{
    const BITMAPINFO2 * bmpInfo2 = renderPage->getBitmapInfoPtr();
    if (bmpInfo2 == NULL)
    {
        throw GbmException(GBM_ERR_BAD_ARG, "Render page has no bitmap header");
    }

    memset(gbmrgb, 0, sizeof(GBMRGB) * 0x100);

    if (renderPage->getColorDepth() > 8)
    {
        memset(bmpHeader, 0, sizeof(BMPHEADER));
        memcpy(bmpHeader, bmpInfo2, sizeof(BITMAPINFOHEADER2));
        *isGrayscale8bpp = FALSE;
    }
    else
    {
        memcpy(bmpHeader, bmpInfo2, sizeof(BMPHEADER));

        const int palette_entries = 1 << renderPage->getColorDepth();
        *isGrayscale8bpp = (palette_entries == 256) ? TRUE : FALSE;

        for (int i = 0; i < palette_entries; i++)
        {
            GBMRGB * pRGB = &(gbmrgb[i]);

            pRGB->r = bmpHeader->argb2Color[i].bRed;
            pRGB->g = bmpHeader->argb2Color[i].bGreen;
            pRGB->b = bmpHeader->argb2Color[i].bBlue;

            if ((pRGB->r != pRGB->g) ||
                (pRGB->r != pRGB->b) ||
                (pRGB->g != pRGB->b))
            {
                *isGrayscale8bpp = FALSE;
            }
        }
    }
}

/* ---------------------------- */

void GbmRenderer::copyGBMRGB2BmpHeader(const GBMRGB    * gbmrgb,
                                             BMPHEADER * bmpHeader) const
{
    for (int i = 0; i < 256; i++)
    {
        bmpHeader->argb2Color[i].bRed   = gbmrgb[i].r;
        bmpHeader->argb2Color[i].bGreen = gbmrgb[i].g;
        bmpHeader->argb2Color[i].bBlue  = gbmrgb[i].b;
    }
}



